<?php
// This file is part of Moodle - https://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <https://www.gnu.org/licenses/>.

/**
 * This class contains the database schema generator for the AAT plugin.
 *
 * @package     local_aatplugin
 * @copyright   2024 Joseph Larochelle <larochellejo@hotmail.com>, Richard Jacklin <rijacklin1@gmail.com>
 * @license     https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

namespace local_aatplugin;

defined('MOODLE_INTERNAL') || die();

/**
 * Class to generate the SQL schema for the AAT plugin.
 */
class database_schema_generator {
    /**
     * Constant representing the integer type.
     */
    const TYPE_INT = 'int';
    /**
     * Constant representing the character type.
     */
    const TYPE_CHAR = 'char';
    /**
     * Constant representing the number type.
     */
    const TYPE_NUMBER = 'number';
    /**
     * Constant representing the bigint SQL data type.
     */
    const BIGINT = 'bigint';
    /**
     * Constant representing the character varying SQL data type.
     */
    const CHAR_VARYING = 'character varying';
    /**
     * Constant representing the numeric SQL data type.
     */
    const NUMERIC = 'numeric';
    /**
     * Constant representing the bigserial SQL data type.
     */
    const BIGSERIAL = 'BIGSERIAL PRIMARY KEY';

    /**
     * Generate SQL CREATE TABLE statement for a given table structure. This
     * function is called from the 'index.php' file. To generate the schema of
     * the moodle database.
     *
     * @param SimpleXMLElement $table The table structure as a SimpleXMLElement.
     * @return string The SQL CREATE TABLE statement.
     */
    public function generatecreatetablesql($table) {
        global $CFG;
        $tablename = $table['NAME'];
        $fieldssql = $this->generatefieldssql($table->FIELDS->FIELD);
        $sql = "CREATE TABLE public.$CFG->prefix$tablename (\n";
        $sql .= implode(",\n", $fieldssql) . "\n);\n\n";
        return $sql;
    }
    /**
     * Generate SQL for table fields.
     *
     * @param SimpleXMLElement[] $fields The fields as an array of SimpleXMLElement objects.
     * @return array An array of strings, each representing the SQL for a field.
     */
    public function generatefieldssql($fields) {
        $fieldssql = [];
        foreach ($fields as $field) {
            $fieldsql = $this->generatefieldsql($field);
            $fieldssql[] = $fieldsql;
        }
        return $fieldssql;
    }

    /**
     * Determine the SQL data type for a field.
     *
     * @param SimpleXMLElement $field The field as a SimpleXMLElement.
     * @return string The SQL data type for the field.
     */
    public function getfieldsqltype($field) {
        switch ($field['TYPE']) {
            case self::TYPE_INT:
                return self::BIGINT;
            case self::TYPE_CHAR:
                return self::CHAR_VARYING;
            case self::TYPE_NUMBER:
                return self::NUMERIC . "(" . $field['LENGTH'] . ", " . $field['DECIMALS'] . ")";
            default:
                return $field['TYPE'];
        }
    }

    /**
     * Generate SQL for a single field.
     *
     * @param SimpleXMLElement $field The field as a SimpleXMLElement object.
     * @return string The SQL for the field.
     */
    public function generatefieldsql($field) {
        $fieldname = $field['NAME'];
        $fieldtype = $this->getfieldsqltype($field);
        $fieldsql = "\t$fieldname $fieldtype";

        // Handle the default value.
        if (!empty($field['DEFAULT'])) {
            $fieldsql .= " DEFAULT '" . $field['DEFAULT'] . "'";
        }

        // Handle NOT NULL constraint.
        if ($field['NOTNULL'] == 'true') {
            $fieldsql .= $field['TYPE'] == self::TYPE_CHAR && empty($field['DEFAULT']) ? " DEFAULT ''::character varying NOT NULL" :
                    " NOT NULL";
        }

        // Handle the sequence for ID fields.
        if ($field['SEQUENCE'] == 'true' && $fieldname == 'id') {
            $fieldsql = "\tid " . self::BIGSERIAL;
        }
        return $fieldsql;
    }
}
