<?php
// This file is part of Moodle - https://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <https://www.gnu.org/licenses/>.

/**
 * Handles plugin file manipulation.
 *
 * @package     local_aatplugin
 * @copyright   2024 Joseph Larochelle <larochellejo@hotmail.com>, Richard Jacklin <rijacklin1@gmail.com>
 * @license     https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

namespace local_aatplugin;

defined('MOODLE_INTERNAL') || die();

class file_handler {
    /**
     * Plugin filesystem component name
     * @var string
     */
    private $component;

    /**
     * Plugin filesystem filearea name
     * @var string
     */
    private $filearea;

    /**
     * Current file's itemid
     * @var int
     */
    private $itemid;

    /**
     * Plugin filesystem context id
     * @var int
     */
    private $contextid;

    /**
     * Constructor for the file_handler class.
     *
     * @param int $itemid (optional) Pass the currentfile's itemid
     */
    public function __construct($itemid = null) {
        global $CFG;
        require_once($CFG->libdir . '/accesslib.php');
        
        try {
            $syscontext = \context_system::instance();
            if (!$syscontext) {
                throw new \moodle_exception('invalidcontext');
            }
            $this->contextid = $syscontext->id;
        } catch (\Exception $e) {
            throw new \moodle_exception('contextinitialization');
        }
        
        $this->component = 'local_aatplugin';
        $this->filearea = 'temp';
        $this->itemid = $itemid ?? time();
    }

	/**
     * Set the item ID for this handler.
	 *
     * @param int $itemid The item ID to use
     */
    public function set_itemid($itemid) {
        $this->itemid = $itemid;
    }

	/**
     * Get the current item ID.
     * 
     * @return int The current item ID
     */
    public function get_itemid() {
		return $this->itemid;
	}

    /**
     * Check the status of the processed SQL data dump file to be downloaded.
     *
     * @param string $filename Name of the file to check status of
     */
    public function check_file_status($filename) {
        $fs = get_file_storage();
        
		$file = $fs->get_file(
			$this->contextid,
			$this->component,
			$this->filearea,
			$this->itemid,
			'/',
			$filename
		);

		if ($file) {
			$url = \moodle_url::make_pluginfile_url(
				$this->contextid,
				$this->component,
				$this->filearea,
				$this->itemid,
				'/',
				$filename,
				true
			);
			$url->param('sesskey', sesskey());

			return [
				'status' => 'complete',
				'url' => $url->out(false),
				'filesize' => $file->get_filesize()
			];
		} else {
			return [
				'status' => 'pending',
				'url' => ''
			];
		}
    }

    /**
     * Handles creating a temporary file using Moodle's File API
     *
     * @param string $content Content to append to created file
     * @param string $filename Name of the file to create
     */
    public function create_temp_file($content, $filename) {
        $fs = get_file_storage();
        
        // Check for existing file first
        $existing_file = $this->get_temp_file($filename);
        if ($existing_file) {
			// If file already exists, delete it first
            $existing_file->delete();
        }
        
		$filerecord = [
			'contextid' => $this->contextid,
			'component' => $this->component,
			'filearea'  => $this->filearea,
			'itemid'    => $this->itemid,
			'filepath'  => '/',
			'filename'  => $filename,
			'timecreated' => time(),
			'timemodified' => time()
		];
		
		return $fs->create_file_from_string($filerecord, $content);
    }

    /**
     * Handles retrieving a temporary file from Moodle plugin storage
     *
     * @param string $filename Name of the file to retrieve
     */
    public function get_temp_file($filename) {
        $fs = get_file_storage();
        return $fs->get_file(
            $this->contextid,
            $this->component,
            $this->filearea,
            $this->itemid,
            '/',
            $filename
        );
    }
}
