<?php
// This file is part of Moodle - https://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <https://www.gnu.org/licenses/>.

/**
 * Form for generating a data dump.
 *
 * @package     local_aatplugin
 * @copyright   2024 Joseph Larochelle <larochellejo@hotmail.com>, Richard Jacklin <rijacklin1@gmail.com>
 * @license     https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

namespace local_aatplugin\form;

defined('MOODLE_INTERNAL') || die();

use renderable;
use renderer_base;
use templatable;
use moodleform;

/**
 * Form for generating a data dump.
 */
class dump_form extends moodleform implements renderable, templatable {
    /**
	 * The courses where the current user is an instructor.
	 * @var array
     */
    private $courses;

    /**
     * Form submission errors.
     * @var array
     */
	private $errors;

    /**
     * Whether to show the download form template.
     * @var bool
     */
	private $show_download_form = false;

    /**
     * Filename of generated SQL dump.
     * @var string
     */
	private $file_name;

    /**
     * ID of data_dump task.
     * @var int
     */
	private $item_id;

    /**
	 * Maximum number of times to retry file check.
	 * @var int
     */
	private $max_retries = 3;

    /**
	 * Delay between retrying file check.
	 * @var int
     */
    private $retry_delay = 2;

    /**
     * Status to SQL dump file.
     * @var array
     */
    private $file_status = null;

    /**
     * Download URL of file.
     * @var moodle_url
     */
	private $download_url = null;

    /**
     * Constructor for the dump_form class.
     *
     * @param bool $show_download_form (optional) Whether to show the download form template
     * @param array $selected_courses (optional) Array of selected course ids
     */
    public function __construct($show_download_form = false, $item_id = null, $file_name = null) {
	 	$this->show_download_form = $show_download_form;
		$this->item_id = $item_id;
        $this->file_name = $file_name;
		$this->courses = local_aatplugin_fetch_user_courses();

		// If we're showing download form, check file status immediately
        if ($show_download_form && $item_id && $file_name) {
            $this->check_file_with_retries();
        }

        parent::__construct();
    }

	private function check_file_with_retries() {
        $handler = new \local_aatplugin\file_handler($this->item_id);
        $attempts = 0;
        
        while ($attempts < $this->max_retries) {
            $file_status = $handler->check_file_status($this->file_name);
            
            if ($file_status['status'] === 'complete') {
                $this->file_status = $file_status;
                
                if (empty($file_status['url'])) {
                    $context = \context_system::instance();
                    $url = \moodle_url::make_pluginfile_url(
                        $context->id,
                        'local_aatplugin',
                        'temp',
                        $this->item_id,
                        '/',
                        $this->file_name,
                        true
                    );
                    $url->param('sesskey', sesskey());
                    $this->download_url = $url->out(false);
                } else {
                    $this->download_url = $file_status['url'];
                }
                return;
            }
            
            $attempts++;
            if ($attempts < $this->max_retries) {
                sleep($this->retry_delay);
            }
        }
        
        $this->show_download_form = false;
        $this->errors[] = ['message' => get_string('filegenerationfailed', 'local_aatplugin')];
    }

    /**
     * Determine which form to generate.
     */
    public function definition() {
		global $OUTPUT;
		$mform = $this->_form;

		if ($this->show_download_form) {
			if ($this->file_status && $this->file_status['status'] === 'complete' && $this->download_url) {
                // Download SQL dump template
                $download_form = new download_form(
                    $this->item_id, 
                    $this->file_name,
                    $this->file_status['url']
                );
                $download = $OUTPUT->render($download_form);
                $mform->addElement('static', 'download', '', $download);
            } else {
                // Add error message and return to course selection
                $this->show_download_form = false;
                $mform->addElement('static', 'error', '', get_string('filegenerationpending', 'local_aatplugin'));
            }
        } else {
			// Create a group for course checkboxes
            $mform->addElement('header', 'courseselection', get_string('selectcourses', 'local_aatplugin'));
            
            // Add checkboxes for each course
            foreach ($this->courses as $course) {
                $mform->addElement('advcheckbox', 
                    'course_' . $course['id'],
                    '',
                    $course['fullname'],
                    array('group' => 'courses'),
                    array(0, $course['id'])
                );
            }

            $mform->addElement('submit', 'generate_dump', get_string('generatedump', 'local_aatplugin'));
        }
    }

    /**
     * Validate form submission.
     *
     * @param array $data Array of submitted form data
     * @param array $files Array of uploaded files
	 *
	 * @return array An array of errors
     */
	public function validation($data, $files) : array {
		// Store any validation errors from parent class
		$errors = parent::validation($data, $files);

        if (!$this->show_download_form) {
            // Check if any course is selected
            $selected = false;
            foreach ($this->courses as $course) {
                if (!empty($data['course_' . $course['id']])) {
                    $selected = true;
                    break;
                }
            }
            
            if (!$selected) {
                $errors['courseselection'] = get_string('nocourseselected', 'local_aatplugin');
            }
        }
 
        return $errors;
	}

	public function get_data() {
		$data = parent::get_data();
        if ($data) {
            // Convert checkbox data to course array
            $selected_courses = [];
            foreach ($this->courses as $course) {
                $field_name = 'course_' . $course['id'];
                if (!empty($data->{$field_name})) {
                    $selected_courses[] = $data->{$field_name};
                }
            }
            $data->selected_courses = $selected_courses;
        }
        return $data;
    }

	/**
	 * Get the value of show_download_form.
	 *
	 * @return bool True if download form should be shown, false otherwise
	 */
	public function get_show_download_form() : bool {
		return $this->show_download_form;
	}

	/**
	 * Set the value of show_download_form.
	 *
	 * @param bool Whether to show the download form template
	 *
	 * @return void
	 */
	public function set_show_download_form($show) : void {
		$this->show_download_form = $show;
	}

	/**
	 * Export data to mustache template for proper output.
	 *
	 * @param renderer_base $output
	 *
	 * @return array Data ready for use in a mustache template
	 */
	public function export_for_template(renderer_base $output) : array {
		$validation = $this->_form->validate();

		if (!$validation) {
			foreach ($this->_form->_errors as $field => $message) {
				$this->errors[] = ['message' => $message];
			}
		}

		return [
			'form' => $this->render(),
			'errors' => $this->errors,
			'courses' => $this->courses,
			'show_download_form' => $this->get_show_download_form(),
			'file_status' => $this->file_status
		];
	}
}
