<?php
// This file is part of Moodle - https://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <https://www.gnu.org/licenses/>.

/**
 * Adhoc task that generates SQL data dump file
 *
 * @package     local_aatplugin
 * @copyright   2024 Joseph Larochelle <larochellejo@hotmail.com>, Richard Jacklin <rijacklin1@gmail.com>
 * @license     https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

namespace local_aatplugin\task;

defined('MOODLE_INTERNAL') || die();

require_once(__DIR__ . '/../../lib.php');

use Exception;

/**
 * Class to generate the SQL dump for the AAT plugin.
 */
class data_dump extends \core\task\adhoc_task {
    private $handler;
    private $buffer = '';
    private $buffer_size = 0;
    private $max_buffer_size = 1048576; // 1MB buffer
    private $total_written = 0;
	private $userids;

    public function execute() {
        global $DB, $CFG;

		$data = $this->get_custom_data();
		
		// Initialize retry count
		$retry_count = isset($data->retry_count) ? $data->retry_count : 0;
		$max_retries = 3;
 
		// Check if we've exceeded max retries
		if ($retry_count >= $max_retries) {
			throw new \moodle_exception('maxretries', 'local_aatplugin');
		}

		// Initialize file context and handler
		$syscontext = \context_system::instance();
		if (!$syscontext) {
			throw new \moodle_exception('invalidcontext');
		}
		$this->handler = new \local_aatplugin\file_handler($data->itemid);

		// Load the public key.
		$publickey = openssl_get_publickey(file_get_contents($CFG->dirroot . '/local/aatplugin/public_key.pem'));
		if (!$publickey) {
			throw new Exception('Failed to load public key');
		}

		$relationsmapper = new \local_aatplugin\coursetable_relations_mapper();
		$databaseschemagenerator = new \local_aatplugin\database_schema_generator();
		$courseids = array_values($data->selected_courses);
		
		// Initialize dump file with headers
		$this->buffer_write("-- AAT plugin SQL dump\n", $data->file_name);
		$this->buffer_write("-- {$CFG->release}\n\n", $data->file_name);

		// Process schema
		$combined_xml_file = $this->ensure_xml_files_exist();
		$this->process_schema($combined_xml_file, $databaseschemagenerator, $data->file_name);
		
		// Process related tables mapping
		$related_tables_file = $this->process_related_tables($combined_xml_file, $relationsmapper);
		$relatedtables = $this->get_related_tables_data($related_tables_file);

		// Initialize course tables array
		$coursetables = $this->initialize_course_tables($relatedtables);

		// Process main course data
		$this->process_course_data($DB, $courseids, $data->file_name, $coursetables);
		
		// Process config data
		$this->process_config_data($DB, $data->file_name, $coursetables);

		// Process user data from logs
		$userids = $this->process_log_data($DB, $courseids);

		// Process user and related data
		$this->process_user_data($DB, $userids, $data->file_name, $coursetables);
		
		// Process question data
		$this->process_question_data($DB, $userids, $data->file_name, $coursetables);

		// Process context data
		$this->process_context_data($DB, $data->file_name, $coursetables);

		// Process related tables data
		$this->process_related_data($DB, $relatedtables, $coursetables, $data->file_name);

		// Process additional tables
		$this->process_additional_tables($DB, $coursetables, $data->file_name);

		// Final flush and verification
		$this->flush_buffer($data->file_name);

		$result = $this->handler->check_file_status($data->file_name);
		if ($result['status'] === 'pending') {
			// Don't delete the file, just schedule retry
			$data->retry_count = isset($data->retry_count) ? $data->retry_count + 1 : 1;
			$new_task = new \local_aatplugin\task\data_dump();
			$new_task->set_custom_data($data);
			$new_task->set_next_run_time(time() + 30); // Retry after 30 seconds
			
			\core\task\manager::queue_adhoc_task($new_task);
			throw new \failed_task_exception("File generation incomplete");
		}
	}

	/**
	 * Process database schema from XML and write to dump file
	 * 
	 * @param stored_file $combined_xml_file The combined XML file
	 * @param object $databaseschemagenerator The schema generator instance
	 * @param string $filename The name of the file to write to
	 * @return stored_file|null The schema file
	 */
	private function process_schema($combined_xml_file, $databaseschemagenerator, $filename) {
		// Check if the schema file exists
		$database_schema_file = $this->handler->get_temp_file('database_schema.sql');
		
		if (!$database_schema_file) {
			// Get XML content from stored_file and load it
			$xml_content = $combined_xml_file->get_content();
			$xml = simplexml_load_string($xml_content);
			if (!$xml) {
				throw new \moodle_exception('invalidxml', 'local_aatplugin');
			}
			
			$sqlstatement = '';
			foreach ($xml->TABLES->TABLE as $table) {
				$sqlstatement .= $databaseschemagenerator->generatecreatetablesql($table);
			}
			
			$database_schema_file = $this->handler->create_temp_file($sqlstatement, 'database_schema.sql');
			
			if (!$database_schema_file) {
				throw new \moodle_exception('xmlfilenotcreated', 'local_aatplugin');
			}
		}

		// Write schema to dump file
		$schema = $database_schema_file->get_content();
		if (!empty($schema)) {
			$this->buffer_write($schema, $filename);
		}
		
		return $database_schema_file;
	}

	/**
	 * Get related tables data from the stored file or array
	 * 
	 * @param stored_file|array $related_tables_file The file or array containing related tables data
	 * @return array The related tables data
	 * @throws moodle_exception If the data cannot be processed
	 */
	private function get_related_tables_data($related_tables_file) {
		if (is_array($related_tables_file)) {
			return $related_tables_file;
		}
		
		if (!$related_tables_file) {
			throw new \moodle_exception('relatedtablesnotfound', 'local_aatplugin');
		}

		$serializedtables = $related_tables_file->get_content();
		$relatedtables = unserialize($serializedtables, ["allowed_classes" => false]);
		
		if ($relatedtables === false) {
			throw new \moodle_exception('unserializefailed', 'local_aatplugin');
		}
		
		return $relatedtables;
	}

	/**
	 * Initialize the course tables array by identifying direct children of the course table
	 * 
	 * @param array $relatedtables Array of tables and their relationships
	 * @return array Array of course tables with empty arrays for IDs
	 */
	private function initialize_course_tables($relatedtables) {
		$coursetables = [];
		
		// Iterate through and collect all tables that do not have a subtable
		// These tables are direct children of the course table
		foreach ($relatedtables as $table => $subtable) {
			// If the table has no subtable, add it to the array of course tables
			if ($subtable == null) {
				$coursetables[$table] = [];
				// Remove it from the array of related tables
				unset($relatedtables[$table]);
			}
		}
		
		// Remove the question table as it's handled separately
		unset($coursetables['question']);
		
		return $coursetables;
	}

	/**
	 * Process and write the 'release' record from the config table
	 * 
	 * @param \moodle_database $DB The Moodle database instance
	 * @param string $filename The name of the file to write to
	 * @param array $coursetables Reference to the course tables array
	 */
	private function process_config_data($DB, $filename, &$coursetables) {
		try {
			// Add only the 'release' record from the mdl_config table to the sql dump
			$sql = "SELECT * FROM {config} WHERE name = ?";
			$params = ['release'];
			$configsrecord = $DB->get_recordset_sql($sql, $params);
			
			$sql = $this->create_sql_statements($configsrecord, 'config', $coursetables);
			if (!empty($sql)) {
				$this->buffer_write($sql, $filename);
			}
		} finally {
			if (isset($configsrecord)) {
				$configsrecord->close();
			}
		}
	}

    /**
     * Process related tables from XML
     */
    private function process_related_tables($combined_xml_file, $relationsmapper) {
        $xml_content = $combined_xml_file->get_content();
        $xml = simplexml_load_string($xml_content);
        
        if (!$xml) {
            throw new \moodle_exception('invalidxml', 'local_aatplugin');
        }

        $initialtables = $relationsmapper->findinitialtables($xml);
        return $relationsmapper->findallrelatedtables($xml, $initialtables);
    }

    /**
     * Process course data and build course tables array
     */
    private function process_course_data($DB, $courseids, $filename, &$coursetables) {
        try {
            $coursesrecords = $DB->get_recordset_list('course', 'id', $courseids);
            $sql = $this->create_sql_statements($coursesrecords, 'course', $coursetables);
            if (!empty($sql)) {
                $this->buffer_write($sql, $filename);
            }
        } finally {
            if (isset($coursesrecords)) {
                $coursesrecords->close();
            }
        }

        // Get all the distinct userids from the logstore_standard_log table where courseid is in the array of courseids.
        $logstorestandardlogrecords = $DB->get_records_list('logstore_standard_log', 'courseid', $courseids);
        $this->userids = [];
        foreach ($logstorestandardlogrecords as $logstorestandardlogrecord) {
            if (!empty($logstorestandardlogrecord->userid)) {
                $this->userids[$logstorestandardlogrecord->userid] = true;
            }
            if (!empty($logstorestandardlogrecord->relateduserid)) {
                $this->userids[$logstorestandardlogrecord->relateduserid] = true;
            }
        }
    }

	/**
     * Process user data from logs
     */
    private function process_log_data($DB, $courseids) {
        $logstorestandardlogrecords = $DB->get_records_list('logstore_standard_log', 'courseid', $courseids);
        $userids = [];
        foreach ($logstorestandardlogrecords as $record) {
            if (!empty($record->userid)) {
                $userids[$record->userid] = true;
            }
            if (!empty($record->relateduserid)) {
                $userids[$record->relateduserid] = true;
            }
        }
        return $userids;
    }

    /**
     * Process user data
     */
    private function process_user_data($DB, $userids, $filename, &$coursetables) {
        if (!empty($userids)) {
            try {
                $usersrecords = $DB->get_recordset_list('user', 'id', array_keys($userids));
                $sql = $this->create_sql_statements($usersrecords, 'user', $coursetables);
                if (!empty($sql)) {
                    $this->buffer_write($sql, $filename);
                }
            } finally {
                if (isset($usersrecords)) {
                    $usersrecords->close();
                }
            }
        }
    }

    /**
     * Process question data
     */
    private function process_question_data($DB, $userids, $filename, &$coursetables) {
        if (!empty($userids)) {
            // Process questions created by users
            try {
                $created_questions = $DB->get_recordset_list('question', 'createdby', array_keys($userids));
                $sql = $this->create_sql_statements($created_questions, 'question', $coursetables);
                if (!empty($sql)) {
                    $this->buffer_write($sql, $filename);
                }
            } finally {
                if (isset($created_questions)) {
                    $created_questions->close();
                }
            }

            // Process questions modified by users
            try {
                $modified_questions = $DB->get_recordset_list('question', 'modifiedby', array_keys($userids));
                $sql = $this->create_sql_statements($modified_questions, 'question', $coursetables);
                if (!empty($sql)) {
                    $this->buffer_write($sql, $filename);
                }
            } finally {
                if (isset($modified_questions)) {
                    $modified_questions->close();
                }
            }
        }
    }

    /**
     * Process context data
     */
    private function process_context_data($DB, $filename, &$coursetables) {
        try {
            $contextsrecords = $DB->get_recordset('context');
            $sql = $this->create_sql_statements($contextsrecords, 'context', $coursetables);
            if (!empty($sql)) {
                $this->buffer_write($sql, $filename);
            }
        } finally {
            if (isset($contextsrecords)) {
                $contextsrecords->close();
            }
        }
    }

	/**
	 * Process additional required tables that aren't directly related to courses
	 * 
	 * @param \moodle_database $DB The Moodle database instance
	 * @param array $coursetables Reference to the course tables array
	 * @param string $filename The name of the file to write to
	 */
	private function process_additional_tables($DB, &$coursetables, $filename) {
		// Book chapters
		if (isset($coursetables['book']) && !empty($coursetables['book'])) {
			try {
				$bookchaptersrecords = $DB->get_recordset_list('book_chapters', 'bookid', array_unique($coursetables['book']));
				$sql = $this->create_sql_statements($bookchaptersrecords, 'book_chapters', $coursetables);
				if (!empty($sql)) {
					$this->buffer_write($sql, $filename);
				}
			} finally {
				if (isset($bookchaptersrecords)) {
					$bookchaptersrecords->close();
				}
			}
		}

		// Course categories
		try {
			$coursecategoriesrecords = $DB->get_recordset('course_categories');
			$sql = $this->create_sql_statements($coursecategoriesrecords, 'course_categories', $coursetables);
			if (!empty($sql)) {
				$this->buffer_write($sql, $filename);
			}
		} finally {
			if (isset($coursecategoriesrecords)) {
				$coursecategoriesrecords->close();
			}
		}

		// Course modules completion
		if (isset($coursetables['course_modules']) && !empty($coursetables['course_modules'])) {
			try {
				$coursemodulescompletionrecords = $DB->get_recordset_list('course_modules_completion', 
					'coursemoduleid', 
					array_unique($coursetables['course_modules']));
				$sql = $this->create_sql_statements($coursemodulescompletionrecords, 'course_modules_completion', $coursetables);
				if (!empty($sql)) {
					$this->buffer_write($sql, $filename);
				}
			} finally {
				if (isset($coursemodulescompletionrecords)) {
					$coursemodulescompletionrecords->close();
				}
			}
		}

		// Files
		if (isset($coursetables['user']) && !empty($coursetables['user'])) {
			try {
				$filesrecords = $DB->get_recordset_list('files', 'userid', array_unique($coursetables['user']));
				$sql = $this->create_sql_statements($filesrecords, 'files', $coursetables);
				if (!empty($sql)) {
					$this->buffer_write($sql, $filename);
				}
			} finally {
				if (isset($filesrecords)) {
					$filesrecords->close();
				}
			}
		}

		// Repository instances
		if (isset($coursetables['user']) && !empty($coursetables['user'])) {
			try {
				$repositoryinstancesrecords = $DB->get_recordset_list('repository_instances', 'userid', 
					array_unique($coursetables['user']));
				$sql = $this->create_sql_statements($repositoryinstancesrecords, 'repository_instances', $coursetables);
				if (!empty($sql)) {
					$this->buffer_write($sql, $filename);
				}
			} finally {
				if (isset($repositoryinstancesrecords)) {
					$repositoryinstancesrecords->close();
				}
			}
		}

		// File references
		if (isset($coursetables['repository_instances']) && !empty($coursetables['repository_instances'])) {
			try {
				$filereferencesrecords = $DB->get_recordset_list('file_references', 'repositoryid', 
					array_unique($coursetables['repository_instances']));
				$sql = $this->create_sql_statements($filereferencesrecords, 'file_references', $coursetables);
				if (!empty($sql)) {
					$this->buffer_write($sql, $filename);
				}
			} finally {
				if (isset($filereferencesrecords)) {
					$filereferencesrecords->close();
				}
			}
		}

		// Forum read
		if (isset($coursetables['forum']) && !empty($coursetables['forum'])) {
			try {
				$forumreadrecords = $DB->get_recordset_list('forum_read', 'forumid', 
					array_unique($coursetables['forum']));
				$sql = $this->create_sql_statements($forumreadrecords, 'forum_read', $coursetables);
				if (!empty($sql)) {
					$this->buffer_write($sql, $filename);
				}
			} finally {
				if (isset($forumreadrecords)) {
					$forumreadrecords->close();
				}
			}
		}

		// Forum track preferences
		if (isset($coursetables['forum']) && !empty($coursetables['forum'])) {
			try {
				$forumtrackprefsrecords = $DB->get_recordset_list('forum_track_prefs', 'forumid', 
					array_unique($coursetables['forum']));
				$sql = $this->create_sql_statements($forumtrackprefsrecords, 'forum_track_prefs', $coursetables);
				if (!empty($sql)) {
					$this->buffer_write($sql, $filename);
				}
			} finally {
				if (isset($forumtrackprefsrecords)) {
					$forumtrackprefsrecords->close();
				}
			}
		}

		// LTI submissions
		if (isset($coursetables['lti']) && !empty($coursetables['lti'])) {
			try {
				$ltisubmissionsrecords = $DB->get_recordset_list('lti_submission', 'ltiid', 
					array_unique($coursetables['lti']));
				$sql = $this->create_sql_statements($ltisubmissionsrecords, 'lti_submission', $coursetables);
				if (!empty($sql)) {
					$this->buffer_write($sql, $filename);
				}
			} finally {
				if (isset($ltisubmissionsrecords)) {
					$ltisubmissionsrecords->close();
				}
			}
		}

		// Wiki synonyms
		if (isset($coursetables['wiki_subwikis']) && !empty($coursetables['wiki_subwikis'])) {
			try {
				$wikisynonymsrecords = $DB->get_recordset_list('wiki_synonyms', 'subwikiid', 
					array_unique($coursetables['wiki_subwikis']));
				$sql = $this->create_sql_statements($wikisynonymsrecords, 'wiki_synonyms', $coursetables);
				if (!empty($sql)) {
					$this->buffer_write($sql, $filename);
				}
			} finally {
				if (isset($wikisynonymsrecords)) {
					$wikisynonymsrecords->close();
				}
			}
		}

		// Roles
		try {
			$rolesrecords = $DB->get_recordset('role');
			$sql = $this->create_sql_statements($rolesrecords, 'role', $coursetables);
			if (!empty($sql)) {
				$this->buffer_write($sql, $filename);
			}
		} finally {
			if (isset($rolesrecords)) {
				$rolesrecords->close();
			}
		}

		// Role assignments
		if (isset($coursetables['user']) && !empty($coursetables['user'])) {
			try {
				$roleassignmentsrecords = $DB->get_recordset_list('role_assignments', 'userid', 
					array_unique($coursetables['user']));
				$sql = $this->create_sql_statements($roleassignmentsrecords, 'role_assignments', $coursetables);
				if (!empty($sql)) {
					$this->buffer_write($sql, $filename);
				}
			} finally {
				if (isset($roleassignmentsrecords)) {
					$roleassignmentsrecords->close();
				}
			}
		}
	}

	/**
	 * Process related data based on relationships
	 * 
	 * @param \moodle_database $DB The Moodle database instance
	 * @param array $related_tables Array of related tables and their relationships
	 * @param array $coursetables Reference to the course tables array
	 * @param string $filename The name of the file to write to
	 */
	private function process_related_data($DB, $related_tables, &$coursetables, $filename) {
		foreach ($related_tables as $table => $subtable) {
			// Skip if subtable is null or empty
			if (empty($subtable)) {
				continue;
			}

			// Get the parent table name and field
			$parent_table = key($subtable);
			$parent_field = $subtable[$parent_table];

			// Check if we have data for the parent table
			if (array_key_exists($parent_table, $coursetables) && !empty($coursetables[$parent_table])) {
				// Skip BBB extension tables
				if (!str_contains($table, 'bbbext_simple')) {
					try {
						$records = $DB->get_recordset_list(
							$table, 
							$parent_field, 
							$coursetables[$parent_table]
						);
						
						$sql = $this->create_sql_statements($records, $table, $coursetables);
						
						if (!empty($sql)) {
							$this->buffer_write($sql, $filename);
						}
					} finally {
						if (isset($records)) {
							$records->close();
						}
					}
				}
				
				// Remove processed table from the array
				unset($related_tables[$table]);
			}
		}
	}

	/**
	 * Ensures XML files exist and returns the combined XML file
	 * 
	 * @return stored_file The combined XML file
	 * @throws moodle_exception if file creation fails
	 */
	private function ensure_xml_files_exist() {
		global $CFG;
		
		$itemid = $this->handler->get_itemid();
		
		// Try to get existing file first
		$combined_xml_file = $this->handler->get_temp_file('combined_XML_files.xml');
		
		// If file doesn't exist, create it
		if (!$combined_xml_file) {
			local_aatplugin_combinexmlfiles($CFG->dirroot, $itemid);
			$combined_xml_file = $this->handler->get_temp_file('combined_XML_files.xml');
			
			if (!$combined_xml_file) {
				throw new \moodle_exception('xmlfilenotcreated', 'local_aatplugin');
			}
		}
		
		return $combined_xml_file;
	}

    /**
     * Create SQL statements for records
     */
    private function create_sql_statements($rs, $table, &$course_tables) {
        global $CFG;
        
        if (!$rs) {
            return '';
        }

        $prefixedtablename = $CFG->prefix . $table;
        $sql = "";

        foreach ($rs as $record) {
            if (!isset($record->id)) {
                continue;
            }
            
            $id = $record->id;
            $course_tables[$table][] = $id;
            $sql .= $this->generate_insert_sql($prefixedtablename, $record) . ";\n";
        }

        return $sql;
    }

    private function generate_insert_sql($table, $record) {
        $fields = array_keys((array)$record);
        $values = array_values((array)$record);

        return sprintf(
            "INSERT INTO %s (%s) VALUES (%s)",
            $table,
            implode(', ', $fields),
            implode(', ', array_map([$this, 'quote_value'], $values))
        );
    }

    private function quote_value($value) {
        if (is_null($value)) return 'NULL';
        if (is_numeric($value)) return $value;
        return "'" . addslashes($value) . "'";
    }

    /**
     * Buffers content for writing
     */
    private function buffer_write($content, $file_name) {
        $this->buffer .= $content;
        $this->buffer_size += strlen($content);

        // Auto-flush if buffer exceeds max size
        if ($this->buffer_size >= $this->max_buffer_size) {
            $this->flush_buffer($file_name);
        }
    }

	/**
     * Flushes buffer to file with encryption
     */
    private function flush_buffer($filename) {
        global $CFG;
        
        if (empty($this->buffer)) {
            return;
        }

		$file = $this->handler->get_temp_file($filename);
		
		// Load public key
		$publickey = openssl_get_publickey(file_get_contents($CFG->dirroot . '/local/aatplugin/public_key.pem'));
		if (!$publickey) {
			throw new Exception('Failed to load public key');
		}

		// Get key size for chunking
		$keydetails = openssl_pkey_get_details($publickey);
		$keysize = $keydetails['bits'] / 8 - 11; // Key size in bytes minus padding
		
		// Encrypt buffer content in chunks
		$chunks = str_split($this->buffer, $keysize);
		$encrypted = '';
		
		foreach ($chunks as $chunk) {
			$encryptedchunk = '';
			if (!openssl_public_encrypt($chunk, $encryptedchunk, $publickey)) {
				throw new Exception('Encryption failed');
			}
			$encrypted .= $encryptedchunk;
		}
		
		if (!$file) {
			$file = $this->handler->create_temp_file($encrypted, $filename);
		}

		// Verify write
		if (!$file || $file->get_filesize() == 0) {
			throw new Exception("Failed to write to file: $filename");
		}

		$this->total_written += $this->buffer_size;
		$this->buffer = '';
		$this->buffer_size = 0;
    }
}
