<?php
// This file is part of Moodle - https://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <https://www.gnu.org/licenses/>.

/**
 * This file contains the library functions for the AAT plugin.
 *
 * @package     local_aatplugin
 * @copyright   2024 Joseph Larochelle <larochellejo@hotmail.com>, Richard Jacklin <rijacklin1@gmail.com>
 * @license     https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

defined('MOODLE_INTERNAL') || die();

/**
 * Insert a link to index.php on the site front page navigation menu if the user has the
 * View course overview report capability.
 *
 * @param navigation_node $frontpage Node representing the front page in the navigation tree.
 */
function local_aatplugin_extend_navigation_frontpage(navigation_node $frontpage) {
    require_login();
    $context = \context_system::instance();

    if (has_capability('report/courseoverview:view', $context)) {
        $frontpage->add(
                get_string('pluginname', 'local_aatplugin'),
                new \moodle_url('/local/aatplugin/index.php'),
                navigation_node::TYPE_CUSTOM
        );
    }
}

/**
 * Insert a link to index.php on the course page navigation menu if the user has the
 * View course overview report capability.
 *
 * @param navigation_node $course Node representing the front page in the navigation tree.
 */
function local_aatplugin_extend_navigation_course(navigation_node $course) {
    require_login();
    global $COURSE;
    $courseid = $COURSE->id;
    $context = \context_course::instance($courseid);

    if (has_capability('report/courseoverview:view', $context)) {
        $course->add(
                get_string('pluginname', 'local_aatplugin'),
                new \moodle_url('/local/aatplugin/index.php'),
                navigation_node::TYPE_CUSTOM
        );
    }
}

/**
 * Combines all 'install.xml' files from a given directory and its subdirectories into one large XML file.
 *
 * @param string $directory The base directory to start searching for 'install.xml' files.
 */
function local_aatplugin_combinexmlfiles(string $directory, int $itemid) {
	$fs = get_file_storage();

    $filerecord = [
        'contextid' =>  \context_system::instance()->id,
        'component' => 'local_aatplugin',
        'filearea' => 'temp',
        'itemid' => $itemid,  // Use the provided itemid
        'filepath' => '/',
        'filename' => 'combined_XML_files.xml'
    ];

    $dom = new DOMDocument('1.0', 'UTF-8');
    $root = $dom->createElement('XMLDB');
    $dom->appendChild($root);

    $tableselement = $dom->createElement('TABLES');
    $root->appendChild($tableselement);

    // Initialize a RecursiveIteratorIterator to iterate through directories and files.
    $iterator = new RecursiveIteratorIterator(
            new RecursiveDirectoryIterator($directory),
            RecursiveIteratorIterator::SELF_FIRST
    );

    // Iterate over each item in the directory and its subdirectories.
    foreach ($iterator as $file) {
        if ($file->isFile() && $file->getFilename() == 'install.xml') {
            // Load the XML content of the 'install.xml' file.
            $xml = simplexml_load_file($file->getPathname());
            // Iterate over each '<TABLE>' element within the file.
            foreach ($xml->TABLES->TABLE as $table) {
                $domtable = dom_import_simplexml($table);
                $import = $dom->importNode($domtable, true);
                $tableselement->appendChild($import);
            }
        }
    }

    // Save the combined XML document to a file in Moodle's temp folder
    $dom->formatOutput = true;
	$file_content = $dom->saveXML();
    
    // Save using file API
    $fs->create_file_from_string($filerecord, $file_content);

	return $itemid;
}

/**
 * Fetches the courses where the current user is has the 'report/courseoverview:view' capability.
 *
 * @return array An array of course objects.
 */
function local_aatplugin_fetch_user_courses(): array {
    global $USER;

    // Specify the capability you want to check.
    $capability = 'report/courseoverview:view';

    // Get the list of courses where the user is enrolled.
    $courses = enrol_get_users_courses($USER->id, true);

    // Prepare an array to hold the course ids and fullnames.
    $coursedetails = [];

    // Loop through the courses and check if the user has the specified capability.
    foreach ($courses as $course) {
        $context = \context_course::instance($course->id);
        if (has_capability($capability, $context, $USER->id)) {
            $coursedetails[] = ['id' => (int)$course->id, 'fullname' => $course->fullname];
        }
    }
    return $coursedetails;
}

/**
 * Generates the SQL dump for selected courses and provides a download link.
 *
 * @param array $selected_courses The IDs of the selected courses.
 */
function local_aatplugin_generate_and_queue_export(array $selected_courses) {
    // Generate consistent filename
    $courseidsstring = implode('_', $selected_courses);
    $file_name = "course_{$courseidsstring}_dump_plugin.sql";

    try {
        // Initialize handler with new or existing itemid
        $handler = new \local_aatplugin\file_handler();
        $itemid = $handler->get_itemid();

        // Create new task
        $dump_task = new \local_aatplugin\task\data_dump();
        $dump_task->set_custom_data([
            'selected_courses' => $selected_courses,
            'file_name' => $file_name,
            'itemid' => $itemid
        ]);

        try {
            $task_id = \core\task\manager::queue_adhoc_task($dump_task);
        } catch (\Exception $e) {
            debugging('Task queue error: ' . $e->getMessage());
            throw $e;
        }

        return [
            'filename' => $file_name,
            'task_id' => $task_id,
            'itemid' => $itemid
        ];
    } catch (\Exception $e) {
        return [
            'error' => $e->getMessage(),
        ];
    }
}

/**
 * File serving handler.
 *
 * @param stdClass $course Course object
 * @param stdClass $cm Course module object
 * @param context $context Context object
 * @param string $filearea File area
 * @param array $args Extra arguments
 * @param bool $forcedownload Whether or not force download
 * @param array $options Additional options
 */
function local_aatplugin_pluginfile($course, $cm, $context, $filearea, $args, $forcedownload, $options = array()) {
    require_login();
    
    if ($filearea !== 'temp') {
        return false;
    }

    $itemid = array_shift($args);
    $filename = array_shift($args);
    
    $fs = get_file_storage();
    $file = $fs->get_file($context->id, 'local_aatplugin', $filearea, $itemid, '/', $filename);
    
    if (!$file) {
        return false;
    }

    // Set options for SQL file
    $options = [
        'filename' => $filename,
        'mimetype' => 'application/sql',
        'forcedownload' => true
    ];

    // Send the file
    send_stored_file($file, 0, 0, $forcedownload, $options);
}
